-- $Id: gfx_lups_manager.lua 4440 2009-04-19 15:36:53Z licho $
--------------------------------------------------------------------------------
--------------------------------------------------------------------------------
--
--  author:  jK
--
--  Copyright (C) 2007,2008.
--  Licensed under the terms of the GNU GPL, v2 or later.
--
--------------------------------------------------------------------------------
--------------------------------------------------------------------------------

function widget:GetInfo()
  return {
    name      = "LupsManager",
    desc      = "",
    author    = "jK",
    date      = "Feb, 2008",
    license   = "GNU GPL, v2 or later",
    layer     = 10,
    enabled   = true,
    handler   = true,
  }
end


include("Configs/lupsFXs.lua")

--------------------------------------------------------------------------------
--------------------------------------------------------------------------------

local function MergeTable(table1,table2)
  local result = {}
  for i,v in pairs(table2) do 
    if (type(v)=='table') then
      result[i] = MergeTable(v,{})
    else
      result[i] = v
    end
  end
  for i,v in pairs(table1) do 
    if (result[i]==nil) then
      if (type(v)=='table') then
        if (type(result[i])~='table') then result[i] = {} end
        result[i] = MergeTable(v,result[i])
      else
        result[i] = v
      end
    end
  end
  return result
end


local function blendColor(c1,c2,mix)
  if (mix>1) then mix=1 end
  local mixInv = 1-mix
  return {
    c1[1]*mixInv + c2[1]*mix,
    c1[2]*mixInv + c2[2]*mix,
    c1[3]*mixInv + c2[3]*mix,
    (c1[4] or 1)*mixInv + (c2[4] or 1)*mix
  }
end


local function blend(a,b,mix)
  if (mix>1) then mix=1 end
  return a*(1-mix) + b*mix
end

--------------------------------------------------------------------------------
--------------------------------------------------------------------------------

local UnitEffects = {}

local function AddFX(unitname,fx)
  local ud = UnitDefNames[unitname]
  if ud then
    UnitEffects[ud.id] = fx
  end
end

--------------------------------------------------------------------------------
--------------------------------------------------------------------------------
--[[AddFX("bomber", {
    {class='AirJet',options={color={0.1,0.4,0.6}, width=2, length=25, piece="flare_r", onActive=true}},
    {class='AirJet',options={color={0.1,0.4,0.6}, width=2, length=25, piece="flare_l", onActive=true}},
	{class='Ribbon',options={color    = {1.0,0.7,0,0.5},width=1, size=30, piece="flare_r"}},
    	{class='Ribbon',options={color    = {1.0,0.7,0,0.5},width=1, size=30, piece="flare_l"}},
	{class='Ribbon',options={color    = {0.0,0.7,1.0,0.5},width=2, size=90, piece="flare_r"}},
    	{class='Ribbon',options={color    = {0.0,0.7,10,0.5},width=2, size=90, piece="flare_l"}},
  })]]


--[[--// FUSIONS //--------------------------
AddFX("cafus", {
    {class='Bursts',options=cafusBursts},
    {class='StaticParticles',options=cafusCorona},
    {class='ShieldSphere',options=cafusShieldSphere},
    {class='ShieldJitter',options={layer=-16, life=math.huge, pos={0,58.9,-4.5}, size=24.5, precision=22, repeatEffect=true}},
    {class='GroundFlash',options=groundFlashBlue},
  })
AddFX("corfus", {
    {class='StaticParticles',options=corfusNova},
    {class='StaticParticles',options=corfusNova2},
    {class='StaticParticles',options=corfusNova3},
    {class='StaticParticles',options=corfusNova4},
    {class='StaticParticles',options=corfusNova5},

    {class='Bursts',options=corfusBursts},
    --{class='StaticParticles',options=corfusCorona},
    --{class='ShieldSphere',options=corfusShieldSphere},
    {class='ShieldJitter',options={delay=30,life=math.huge, pos={0,40.5,0}, size=22, precision=22, repeatEffect=true}},
  })

--// SHIELDS //---------------------------
AddFX("corjamt", {
    {class='Bursts',options=corjamtBursts},
    {class='ShieldSphere',options={life=math.huge, pos={3.5,30,0}, size=11, colormap1 = {{0.8, 0.0, 0.8, 0.8}}, repeatEffect=true}},
    {class='GroundFlash',options=groundFlashViolett},
  })
AddFX("coreter", {
    {class='Bursts',options=coreterBursts},
    {class='ShieldSphere',options={piece="sphere", life=math.huge, size=8, colormap1 = {{0.95, 0.1, 0.95, 0.9}}, repeatEffect=true}},
  })
AddFX("core_spectre", {
    {class='Bursts',options=coreterBursts},
    {class='ShieldSphere',options={piece="sphere", life=math.huge, size=8, colormap1 = {{0.95, 0.1, 0.95, 0.9}}, repeatEffect=true}},
  })
--AddFX("corsjam", {
--  {class='Bursts',options=coreterBursts},
--  {class='ShieldSphere',options={piece="sphere", life=math.huge, size=8, colormap1 = {{0.95, 0.1, 0.95, 0.9}}, repeatEffect=true}},
--})

--// ENERGY STORAGE //--------------------
AddFX("corestor", {
    {class='GroundFlash',options=groundFlashCorestor},
  })
AddFX("armestor", {
    {class='GroundFlash',options=groundFlashArmestor},
  })

AddFX("roost", {
    {class='SimpleParticles',options=roostDirt},
    {class='SimpleParticles',options=MergeTable({delay=60},roostDirt)},
    {class='SimpleParticles',options=MergeTable({delay=120},roostDirt)},
  })

--// OTHER
--AddFX("corgate", {
--    {class='ShieldJitter',options={life=math.huge, pos={0,0,0}, size=300, precision=22, repeatEffect=true}},
--})
AddFX("corarad", {
    {class='StaticParticles',options=radarBlink},
    {class='StaticParticles',options=MergeTable(radarBlink,{pos={-1.6,25,0},delay=15})},
    {class='StaticParticles',options=MergeTable(radarBlink,{pos={0,21,-1.0},delay=30})},
  })
AddFX("corrad", {
    {class='StaticParticles',options=MergeTable(radarBlink,{piece="head"})},
    {class='StaticParticles',options=MergeTable(radarBlink,{piece="head", delay=15})},
  })

--// GROUND UNITS //----------------------
--AddFX("armcrabe", { --// needs to be done in synced lups someday (cus of inAirLos := true)
--  {class='StaticParticles',options=ArmCrabeWhiteLight},
--})
AddFX("armmex", {
    {class='AirJet',options={color={1,0.1,0.1}, width=1, length=30, piece="emit", onActive=false, emitVector={0,1,0}}},
  })


--// SEA PLANES //----------------------------
AddFX("armsaber", {
    {class='AirJet',options={color={0.1,0.4,0.6}, width=3.7, length=25, piece="nozzle", onActive=true}},
  })
AddFX("armseap", {
    {class='AirJet',options={color={0.1,0.4,0.6}, width=3.7, length=30, piece="nozzle", onActive=true}},
  })
AddFX("armsehak", {
    {class='AirJet',options={color={0.1,0.4,0.6}, width=3.7, length=30, piece="nozzle", onActive=true}},
  })
AddFX("armsfig", {
    {class='AirJet',options={color={0.1,0.4,0.6}, width=3.3, length=30, piece="nozzle", onActive=true}},
  })
AddFX("armsb", {
    {class='AirJet',options={color={0.1,0.4,0.6}, width=3.7, length=70, piece="emit1", onActive=true}},
  })
AddFX("corsfig", {
    {class='AirJet',options={color={0.1,0.4,0.6}, width=3.5, length=32, piece="nozzle", onActive=true}},
  })
AddFX("corhunt", {
    {class='AirJet',options={color={0.1,0.4,0.6}, width=6.25, length=40, piece="nozzle", onActive=true}},
  })
AddFX("corseap", {
    {class='AirJet',options={color={0.1,0.4,0.6}, width=3.5, length=25, piece="nozzle1", onActive=true}},
    {class='AirJet',options={color={0.1,0.4,0.6}, width=3.5, length=25, piece="nozzle2", onActive=true}},
  })
AddFX("corsb", {
    {class='AirJet',options={color={0.1,0.4,0.6}, width=3.5, length=36, piece="emit1", onActive=true}},
    {class='AirJet',options={color={0.1,0.4,0.6}, width=3.5, length=36, piece="emit2", onActive=true}},
  })
AddFX("corcut", {
    {class='AirJet',options={color={0.6,0.1,0}, width=6, length=25, piece="rthrust1", onActive=true}},
    {class='AirJet',options={color={0.6,0.1,0}, width=6, length=25, piece="rthrust2", onActive=true}},
  })

--// PLANES //----------------------------
AddFX("armfig", {
    {class='AirJet',options={color={0.3,0.2,0}, width=5.5, length=55, piece="rearthrustbox1", onActive=true}},
    {class='AirJet',options={color={0.3,0.2,0}, width=5.5, length=55, piece="rearthrustbox2", onActive=true}},

    {class='Ribbon',options={piece="wingtip1"}},
    {class='Ribbon',options={piece="wingtip2"}},
    --{class='Ribbon',options={color={1,1,0.6,1}, width=5.5, piece="rearthrustbox1"}},
    --{class='Ribbon',options={color={1,1,0.6,1}, width=5.5, piece="rearthrustbox2"}},
  })
AddFX("armhawk", {
    {class='AirJet',options={color={0.5,0.1,0}, width=5, length=55, piece="rearthrust", onActive=true}},
    {class='Ribbon',options={width=1, size=8, piece="wingtip1"}},
    {class='Ribbon',options={width=1, size=8, piece="wingtip2"}},
  })
AddFX("armcybr", {
    {class='AirJet',options={color={0.5,0.1,0}, width=3.5, length=25, piece="nozzle1", onActive=true}},
	{class='AirJet',options={color={0.5,0.1,0}, width=3.5, length=25, piece="nozzle2", onActive=true}},
   })
AddFX("armhawk2", {
    {class='AirJet',options={color={0.6,0.2,0}, width=2.8, length=12, piece="enginel", onActive=true}},
    {class='AirJet',options={color={0.6,0.2,0}, width=2.8, length=12, piece="enginer", onActive=true}},
    {class='Ribbon',options={width=1, size=12, piece="wingtip1"}},
    {class='Ribbon',options={width=1, size=12, piece="wingtip2"}},
  })
AddFX("armbrawl", {
    {class='AirJet',options={color={0.6,0.2,0}, width=2.8, length=12, piece="nozzle1", onActive=true}},
    {class='AirJet',options={color={0.6,0.2,0}, width=2.8, length=12, piece="nozzle2", onActive=true}},
  })
AddFX("armawac", {
    {class='AirJet',options={color={0.1,0.4,0.6}, width=3.5, length=25, piece="exhaustl", onActive=true}},
    {class='AirJet',options={color={0.1,0.4,0.6}, width=3.5, length=25, piece="exhaustr", onActive=true}},
    {class='Ribbon',options={width=1, size=12, piece="wingtip1"}},
    {class='Ribbon',options={width=1, size=12, piece="wingtip2"}},
  })
AddFX("corgripn", {
    {class='AirJet',options={color={0.5,0.1,0}, width=3.5, length=35, piece="nozzle", onActive=true}},
  })

AddFX("armstiletto_laser", {
    {class='AirJet',options={color={0.1,0.4,0.6}, width=3.5, length=17, piece="jet1", onActive=true}},
	{class='AirJet',options={color={0.1,0.4,0.6}, width=3.5, length=17, piece="jet2", onActive=true}},
    {class='Ribbon',options={width=1, size=6, piece="wingtip1"}},
    {class='Ribbon',options={width=1, size=6, piece="wingtip2"}},
  })


  AddFX("bladew", {
    {class='Ribbon',options={width=1, size=5, piece="ljet"}},
    {class='Ribbon',options={width=1, size=5, piece="rjet"}},	
	{class='AirJet',options={color={0.1,0.4,0.6}, width=3, length=14, piece="ljet", onActive=true, emitVector = {0, 1, 0}}},
	{class='AirJet',options={color={0.1,0.4,0.6}, width=3, length=14, piece="rjet", onActive=true, emitVector = {0, 1, 0}}},
  })

AddFX("armkam", {
    {class='Ribbon',options={width=1, size=10, piece="lfx"}},
    {class='Ribbon',options={width=1, size=10, piece="rfx"}},	
	{class='AirJet',options={color={0.1,0.4,0.6}, width=4, length=25, piece="lfx", onActive=true, emitVector = {0, 0, 1}}},
	{class='AirJet',options={color={0.1,0.4,0.6}, width=4, length=25, piece="rfx", onActive=true, emitVector = {0, 0, 1}}},
  })
  
AddFX("armpnix", {
    {class='AirJet',options={color={0.1,0.4,0.6}, width=3.5, length=25, piece="nozzle", onActive=true}},
  })
AddFX("armthund", {
    {class='AirJet',options={color={0.1,0.4,0.6}, width=3.5, length=15, piece="nozzle1", onActive=true}},
	{class='AirJet',options={color={0.1,0.4,0.6}, width=3.5, length=15, piece="nozzle2", onActive=true}},
  })

--[UnitDefNames["armcybr", {
 --  {class='AirJet',options={color={0.6,0.1,0}, width=3.5, length=22, piece="thrust1", texture2="bitmaps/gpl/lups/jet2.bmp", onActive=true}},
--  {class='AirJet',options={color={0.6,0.1,0}, width=3.5, length=22, piece="thrust2", texture2="bitmaps/gpl/lups/jet2.bmp", onActive=true}},
--},
AddFX("armdfly", {
    {class='AirJet',options={color={0.1,0.5,0.3}, width=3.2, length=22, piece="jet1", onActive=true}},
    {class='AirJet',options={color={0.1,0.5,0.3}, width=3.2, length=22, piece="jet2", onActive=true}},
  })

AddFX("corshad", {
    {class='AirJet',options={color={0.1,0.4,0.6}, width=3.5, length=22, piece="thrustr", onActive=true}},
    {class='AirJet',options={color={0.1,0.4,0.6}, width=3.5, length=22, piece="thrustl", onActive=true}},
  })
AddFX("fighter", {
    {class='AirJet',options={color={0.6,0.1,0}, width=3.5, length=55, piece="nozzle1", onActive=true}},
    {class='AirJet',options={color={0.6,0.1,0}, width=3.5, length=55, piece="nozzle2", onActive=true}},
    {class='Ribbon',options={width=1, piece="wingtip1"}},
    {class='Ribbon',options={width=1, piece="wingtip2"}},
  })
AddFX("corape", {
    {class='AirJet',options={color={0.6,0.1,0}, width=3.5, length=22, piece="rthrust1", onActive=true}},
    {class='AirJet',options={color={0.6,0.1,0}, width=3.5, length=22, piece="rthrust2", onActive=true}},
  })
AddFX("corhurc", {
    {class='AirJet',options={color={0.5,0.1,0}, width=3.5, length=25, piece="nozzle1", onActive=true}},
    {class='AirJet',options={color={0.5,0.1,0}, width=3.5, length=25, piece="nozzle2", onActive=true}},
   })
AddFX("corhurc2", {
    {class='AirJet',options={color={0.1,0.4,0.6}, width=3.5, length=25, piece="rthrust", onActive=true}},
    {class='Ribbon',options={width=1, piece="wingtip1"}},
    {class='Ribbon',options={width=1, piece="wingtip2"}},
  })
AddFX("corvamp", {
    {class='AirJet',options={color={0.6,0.1,0}, width=3.5, length=55, piece="thrustb", onActive=true}},
    {class='Ribbon',options={width=1, size=8, piece="wingtipa"}},
    {class='Ribbon',options={width=1, size=8, piece="wingtipb"}},
  })
AddFX("corawac", {
    {class='AirJet',options={color={0.1,0.4,0.6}, width=3.5, length=25, piece="thrust", onActive=true}},
    --{class='Ribbon',options={color={1,1,0.6,1}, width=4.5, size=9, piece="thrust"}},
    {class='Ribbon',options={width=1, size=8, piece="wingtip1"}},
    {class='Ribbon',options={width=1, size=8, piece="wingtip2"}},
  })
AddFX("blackdawn", {
    {class='AirJet',options={color={0.6,0.2,0}, width=2.8, length=12, piece="nozzle1", onActive=true}},
    {class='AirJet',options={color={0.6,0.2,0}, width=2.8, length=12, piece="nozzle2", onActive=true}},
  })

AddFX("corcrw", {
    {class='AirJet',options={color={0.6,0.15,0}, width=4, length=17, piece="engine", onActive=true}},

    {class='AirJet',options={color={0.5,0.05,0}, width=3.5, length=19, piece="leftengine1", onActive=true}},
    {class='AirJet',options={color={0.5,0.05,0}, width=3.5, length=16, piece="leftengine2", onActive=true}},
    {class='AirJet',options={color={0.5,0.05,0}, width=3.5, length=13, piece="leftengine3", onActive=true}},

    {class='AirJet',options={color={0.5,0.05,0}, width=3.5, length=19, piece="rightengine1", onActive=true}},
    {class='AirJet',options={color={0.5,0.05,0}, width=3.5, length=16, piece="rightengine2", onActive=true}},
    {class='AirJet',options={color={0.5,0.05,0}, width=3.5, length=13, piece="rightengine3", onActive=true}},
  })]]



--------------------------------------------------------------------------------
--------------------------------------------------------------------------------

--[[--// armmex overdrive FX
local armmexDefID = UnitDefNames["armmex"].id
local armmexes    = {}
local armmexesFxIDs = {}
local armmexFX    = armmexJet

--// cormex overdrive FX
local cormexDefID = UnitDefNames["cormex"].id
local cormexes    = {}
local cormexFX    = cormexGlow]]

--------------------------------------------------------------------------------
--------------------------------------------------------------------------------

local abs = math.abs
local min = math.min
local max = math.max
local spGetSpectatingState = Spring.GetSpectatingState
local spGetUnitDefID       = Spring.GetUnitDefID
local spGetUnitRulesParam  = Spring.GetUnitRulesParam

--------------------------------------------------------------------------------
--------------------------------------------------------------------------------

local Lups  -- Lua Particle System
local LupsAddFX
local particleIDs = {}
local initialized = false --// if LUPS isn't started yet, we try it once a gameframe later
local tryloading  = 1     --// try to activate lups if it isn't found

--------------------------------------------------------------------------------
--------------------------------------------------------------------------------

local function ClearFxs(unitID)
  if (particleIDs[unitID]) then
    for _,fxID in ipairs(particleIDs[unitID]) do
      Lups.RemoveParticles(fxID)
    end
    particleIDs[unitID] = nil
  end
end

local function ClearFx(unitID, fxIDtoDel)
  if (particleIDs[unitID]) then
	local newTable = {}
	for _,fxID in ipairs(particleIDs[unitID]) do
		if fxID == fxIDtoDel then 
			Lups.RemoveParticles(fxID)
		else 
			newTable[#newTable+1] = fxID
		end
    end
	if #newTable == 0 then 
		particleIDs[unitID] = nil
	else 
		particleIDs[unitID] = newTable
	end
  end
end



local function AddFxs(unitID,fxID)
  if (not particleIDs[unitID]) then
    particleIDs[unitID] = {}
  end

  local unitFXs = particleIDs[unitID]
  unitFXs[#unitFXs+1] = fxID
end

--------------------------------------------------------------------------------
--------------------------------------------------------------------------------

function widget:Shutdown()
  if (initialized) then
    for _,unitFxIDs in pairs(particleIDs) do
      for _,fxID in ipairs(unitFxIDs) do
        Lups.RemoveParticles(fxID)
      end
    end
    particleIDs = {}
  end

  Spring.SendLuaRulesMsg("lups shutdown","allies")
end

--------------------------------------------------------------------------------
--------------------------------------------------------------------------------
local function UnitFinished(_,unitID,unitDefID)
  --[[if (unitDefID == armmexDefID) then
    armmexes[unitID] = 0
    armmexFX.unit    = unitID
    particleIDs[unitID] = {}
    
	local fxID = LupsAddFX("airjet",armmexFX)
	armmexesFxIDs[unitID] = fxID
	AddFxs( unitID, fxID )
  end

  if (unitDefID == cormexDefID) then
    cormexes[unitID] = 0
    cormexFX.unit    = unitID
    particleIDs[unitID] = {}
    AddFxs( unitID, LupsAddFX("StaticParticles",cormexFX) )
  end]]

  local effects = UnitEffects[unitDefID]
  if (effects) then
    for _,fx in ipairs(effects) do
      if (fx.class=="GroundFlash") then
        fx.options.pos = { Spring.GetUnitBasePosition(unitID) }
      end
      fx.options.unit = unitID
      AddFxs( unitID,LupsAddFX(fx.class,fx.options) )
      fx.options.unit = nil
    end
  end
end

local function UnitDestroyed(_,unitID,unitDefID)
  --[[if (unitDefID == armmexDefID)or(unitDefID == cormexDefID)  then
    armmexes[unitID] = nil
    cormexes[unitID] = nil
  end]]

  ClearFxs(unitID)
end


local function UnitEnteredLos(_,unitID)
  local spec, fullSpec = spGetSpectatingState()
  if (spec and fullSpec) then return end

  --[[if (unitDefID == armmexDefID) then
    armmexes[unitID] = 1
    armmexFX.unit    = unitID
    particleIDs[unitID] = {}
	local fxID = LupsAddFX("airjet",armmexFX)
	armmexesFxIDs[unitID] = fxID
    AddFxs( unitID, fxID )
  end

  if (unitDefID == cormexDefID) then
    cormexes[unitID] = 1
    cormexFX.unit    = unitID
    particleIDs[unitID] = {}
    AddFxs( unitID, LupsAddFX("StaticParticles",cormexFX) )
  end]]

  local unitDefID = spGetUnitDefID(unitID)
  local effects   = UnitEffects[unitDefID]
  if (effects) then
    for _,fx in ipairs(effects) do
      if (fx.class=="GroundFlash") then
        fx.options.pos = { Spring.GetUnitBasePosition(unitID) }
      end
      fx.options.unit = unitID
      AddFxs( unitID,LupsAddFX(fx.class,fx.options) )
      fx.options.unit = nil
    end
  end
end


local function UnitLeftLos(_,unitID)
  local spec, fullSpec = spGetSpectatingState()
  if (spec and fullSpec) then return end

 --[[ if (unitDefID == armmexDefID)or(unitDefID == cormexDefID)  then
    armmexes[unitID] = nil
    cormexes[unitID] = nil
  end]]

  ClearFxs(unitID)
end

--------------------------------------------------------------------------------
--------------------------------------------------------------------------------
local color1 = {0,0,0}
local color2 = {1,0.5,0}

local function GameFrame(_,n)
--[[  if (((n+48)%60)<1 and (next(armmexes) or next(cormexes))) then
    --//Update Overdrive Fx
    for unitID,strength in pairs(armmexes) do
      local cur_strength = spGetUnitRulesParam(unitID,"overdrive") or 1
      local diff         = abs(cur_strength - strength)
      if (diff>0.3) then
        -- limit the maximum change per update (else the fx would jump like hell)
        cur_strength = strength + ((cur_strength>strength and 1) or -1)*0.3

        ClearFx(unitID, armmexesFxIDs[unitID])
		
        armmexFX.unit   = unitID
        armmexFX.color  = blendColor(color1,color2, (cur_strength-1)*0.5)
        armmexFX.length = blend(140,80, (cur_strength-1)*0.75)
        
		local fxID = LupsAddFX("airjet",armmexFX)
		armmexesFxIDs[unitID] = fxID
		AddFxs( unitID, fxID )
        armmexes[unitID] = cur_strength
      end
    end
    armmexFX.color  = color1
    armmexFX.length = 140

    for unitID,strength in pairs(cormexes) do
      local cur_strength = spGetUnitRulesParam(unitID,"overdrive") or 1
      local diff         = abs(cur_strength - strength)
      if (diff>0.1) then
        -- limit the maximum change per update (else the fx would jump like hell)
        cur_strength = (strength) + ((cur_strength>strength and 1) or -1)*0.3

	local a = min(1,max(0,(cur_strength-1)*0.35));
        ClearFxs(unitID)
        cormexFX.unit     = unitID
        cormexFX.colormap = {blendColor(cormexFX.color1,cormexFX.color2, a)}

        cormexFX.size     = blend(cormexFX.size1,cormexFX.size2, a)
        AddFxs( unitID, LupsAddFX("StaticParticles",cormexFX) )
        cormexes[unitID]  = cur_strength
      end
    end
    cormexFX.colormap = {cormexFX.color1}
    cormexFX.size   = cormexFX.size1
  end]]
end

--------------------------------------------------------------------------------
--------------------------------------------------------------------------------

function widget:Update()
  if (Spring.GetGameFrame()<1) then 
    return
  end

  Lups  = WG['Lups']
  local LupsWidget = widgetHandler.knownWidgets['Lups'] or {}

  --// Lups running?
  if (Lups and LupsWidget.active) then
    if (tryloading==-1) then
      Spring.Echo("LuaParticleSystem (Lups) activated.")
    end
    initialized=true
  else
    if (tryloading==1) then
      Spring.Echo("Lups not found! Trying to activate it.")
      widgetHandler:EnableWidget("Lups")
      tryloading=-1
      return
    else
      Spring.Echo("LuaParticleSystem (Lups) couldn't be loaded!")
      widgetHandler:RemoveWidgetCallIn("Update",self)
      return
    end
  end

  --// send errorlog if me (jK) is in the game
  local allPlayers = Spring.GetPlayerList()
  for i=1,#allPlayers do
    local playerName = Spring.GetPlayerInfo(allPlayers[i])
    if (playerName == "[LCC]jK") then
      local errorLog = Lups.GetErrorLog(1)
      if (errorLog~="") then
        local cmds = {
          "say ------------------------------------------------------",
          "say LUPS: jK is here! Sending error log (so he can fix your problems):",
        }
        --// the str length is limited with "say ...", so we split it
        for line in errorLog:gmatch("[^\r\n]+") do
          cmds[#cmds+1] = "say " .. line
        end
        cmds[#cmds+1] = "say ------------------------------------------------------"
        Spring.SendCommands(cmds)
      end
      break
    end
  end

  LupsAddFX = Lups.AddParticles

  Spring.SendLuaRulesMsg("lups running","allies")

  --// initialize effects for existing units
  local allUnits = Spring.GetAllUnits(); allUnits.n = nil;
  for _,unitID in ipairs(allUnits) do
    local unitDefID = Spring.GetUnitDefID(unitID)
    UnitFinished(nil,unitID,unitDefID)
  end

  self.UnitFinished   = UnitFinished
  self.UnitDestroyed  = UnitDestroyed
  self.UnitEnteredLos = UnitEnteredLos
  self.UnitLeftLos    = UnitLeftLos
  self.GameFrame      = GameFrame
  widgetHandler:UpdateWidgetCallIn("UnitFinished",self)
  widgetHandler:UpdateWidgetCallIn("UnitDestroyed",self)
  widgetHandler:UpdateWidgetCallIn("UnitEnteredLos",self)
  widgetHandler:UpdateWidgetCallIn("UnitLeftLos",self)
  widgetHandler:UpdateWidgetCallIn("GameFrame",self)

  widgetHandler:RemoveWidgetCallIn("Update",self)
end

--------------------------------------------------------------------------------
--------------------------------------------------------------------------------
